/*
    QGL Framebuffer
    Ivan Pelipenko peri4ko@yandex.ru

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GLFRAMEBUFFER_H
#define GLFRAMEBUFFER_H

#include "glbuffer.h"


class QGLENGINE_CORE_EXPORT Framebuffer {
	friend class FramebufferMipmap;

public:
	Framebuffer(QOpenGLExtraFunctions * f_,
	            int colorAttachments = 1,
	            bool withDepth       = true,
	            GLenum colorFormat   = GL_RGBA8,
	            GLenum _target       = GL_TEXTURE_2D);
	Framebuffer(QOpenGLExtraFunctions * f_, QVector<GLenum> colors_, bool withDepth = true, GLenum _target = GL_TEXTURE_2D);
	virtual ~Framebuffer();

	GLuint id() const { return fbo; }
	GLuint colorTexture(int index = 0) const { return colors[index]; }
	GLuint depthTexture() const { return tex_d; }
	GLenum target() const { return target_; }
	bool isInit() const { return fbo != 0; }
	int width() const { return wid; }
	int height() const { return hei; }
	QSize size() const { return QSize(wid, hei); }
	QRect rect() const { return QRect(0, 0, wid, hei); }
	QImage grab() const;
	QVector<float> grabF(int index) const;
	void queryPoint(int index, QPoint p);
	void queryPoints(int index, QRect rect, GLenum pixel_format = GL_UNSIGNED_BYTE);
	void queryImage(int index);
	uint getPoint() const;
	QVector<uint> getPointsByte() const;
	QVector<QVector4D> getPointsFloat() const;
	QImage getImage() const;
	int queriedPoints() const { return pbo_queried; }
	void blit(int index_from,
	          GLuint fb_to,
	          int index_to,
	          QRect from,
	          QRect to,
	          GLbitfield mask = GL_COLOR_BUFFER_BIT,
	          GLenum filter   = GL_NEAREST) const;

	void resize(QSize sz, bool force = false) { resize(sz.width(), sz.height(), force); }
	void resize(int width, int height, bool force = false);
	void reinit(QOpenGLExtraFunctions * f_ = nullptr);
	void bind();
	void release();
	void setReadBuffer(int index) { glReadBuffer(GL_COLOR_ATTACHMENT0 + index); }
	void setWriteBuffer(int index);
	void setWriteBuffers(const int * indeces, int count);
	void setWriteBuffers(const QVector<int> & indeces) { setWriteBuffers(indeces.constData(), indeces.size()); }
	void setWriteBuffers();
	void unsetWriteBuffers();
	void enablePixelBuffer();
	void setColorTextureFiltering(int index, GLenum filter);

	void copyDepthFrom(GLuint tex) { ; }
	void bindColorTexture(int index, int channel = 0);
	void bindColorTextures(const QVector<int> & indeces);
	void bindColorTextures();
	void bindDepthTexture(int channel);

private:
	void deleteGLRenderbuffer(GLuint & drbo);
	void deleteGLFramebuffer(GLuint & fbo);

	bool is_depth, is_changed;
	int pbo_queried;
	QOpenGLExtraFunctions * f;
	mutable Buffer pbo;
	QVector<GLuint> colors;
	QVector<GLenum> color_formats;
	GLenum target_;
	GLuint fbo, drbo, tex_d;
	GLint prev_view[4], wid, hei;
};

#endif // GLFRAMEBUFFER_H
